/**
 *--------------------------------------------------------------------\n
 *          HSLU T&A Hochschule Luzern Technik+Architektur            \n
 *--------------------------------------------------------------------\n
 *
 * \brief         Color sensor TCS3414 driver
 * \file
 * \author        Christian Jost, christian.jost@hslu.ch
 * \date          26.11.2012
 *
 * \b Language:   Ansi-C \n\n
 * \b Target:     MC-Car  \n
 *
 * \par Description:
 *
 * $Id: colSens.c 453 2012-12-18 09:44:13Z zajost $
 *--------------------------------------------------------------------
 */
#include <string.h>
#include <stdlib.h>
#include "platform.h"
#include "colSens.h"
#include "i2c.h"
#include "util.h"
#include "term.h"
#include "sci.h"

#define I2C_COLSENS_ADR                 0x72

// TCS3414 register addresses
#define REG_ADR_CONTROL                 0x00
#define REG_ADR_TIMING                  0x01
#define REG_ADR_INT_CTRL                0x02
#define REG_ADR_INT_SOURCE              0x03
#define REG_ADR_ID                      0x04
#define REG_ADR_GAIN                    0x07
#define REG_ADR_LOW_THRESH_LOW_BYTE     0x08
#define REG_ADR_LOW_THRESH_HIGH_BYTE    0x09
#define REG_ADR_HIGH_THRESH_LOW_BYTE    0x0A
#define REG_ADR_HIGH_THRESH_HIGH_BYTE   0x0B
#define REG_ADR_DATA_GREEN_LOW          0x10
#define REG_ADR_DATA_GREEN_HIGH         0x11
#define REG_ADR_DATA_RED_LOW            0x12
#define REG_ADR_DATA_RED_HIGH           0x13
#define REG_ADR_DATA_BLUE_LOW           0x14
#define REG_ADR_DATA_BLUE_HIGH          0x15
#define REG_ADR_DATA_CLEAR_LOW          0x16
#define REG_ADR_DATA_CLEAR_HIGH         0x17

#define REG_ADR_LOW_THRESH              0x08
#define REG_ADR_HIGH_THRESH             0x0A
#define REG_ADR_DATA_GREEN              0x10
#define REG_ADR_DATA_RED                0x12
#define REG_ADR_DATA_BLUE               0x14
#define REG_ADR_DATA_CLEAR              0x16

// Transaction Type
#define REG_CMD_TRANS_BYTE_PROT         0b00
#define REG_CMD_TRANS_WORD_PROT         0b01
#define REG_CMD_TRANS_BLOCK_PROT        0b10
#define REG_CMD_TRANS_INTERRUPT_CLR     0b11


typedef union                               // Register 0x00: Command Register
{
  uint8 Byte;
  struct 
  {
    uint8 Address       :5;   // Bit0       // RegisterAdress
    uint8 Transaction   :2;                 // 00=ByteProtocol, 01=WordProtocol
                                            // 10=BlockProtocol, 11=InteruptClear 
    uint8 CMD           :1;   // Bit7       // must be 1
  } Bits;
} tCommandReg;

typedef union                               // Register 0x00: Control Register
{
  uint8 Byte;
  struct 
  {
    uint8 Power     :1;                     // 1 = power on, 0 = power off r/w
    uint8 AdcEn     :1;                     // 1 = Enable ADC to begin integration r/w
    uint8 Resv0     :1;                     // Reserved 
    uint8 Resv1     :1;                     // Reserved 
    uint8 AdcValid  :1;                     // 1 = ADC value valid, ADC integration cycle completed r 
    uint8 Resv2     :1;                     // Reserved 
    uint8 Resv3     :1;                     // Reserved 
    uint8 Resv4     :1;                     // Reserved 
  } Bits;
} tControlStatusReg;

typedef union                               // Register 0x07: Gain Register
{
  uint8 Byte;
  struct 
  {
    uint8 Prescaler :3;                     // Prescaler setting
    uint8 Resv1     :1;                     // Reserved 
    uint8 Gain      :2;                     // Gain setting
    uint8 Resv2     :2;                     // Reserved
  } Bits;
} tControlGainReg;

typedef struct
{
  uint16 green;
  uint16 red;
  uint16 blue;  
  uint16 clear;
} tColorRawRGB;

static tControlGainReg gainReg;
static tColorRawRGB blackValue;
static tColorRawRGB whiteValue;


/**
 * Swaps the order of the high- and low-byte in a int number.
 * 
 * @param [in|out] value
 *   the value to change the order
 */
void swapByteOrder(uint16 *value)
{
  char* byte1;
  char* byte2;
  int8 tmp;
  
  byte1 = (char *)value;
  byte2 = ((char *)value) + 1;
  
  tmp = *byte1;
  *byte1 = *byte2;
  *byte2 = tmp;
}


/**
 * Reads the raw 16-bit RGB color values from the sensor TCS3414 
 * in a single IIC-Read transfer.
 * (Transaction Type = Block protocol)
 *  
 * @param [out] pColor
 *   pColor->red 
 *   pColor->green 
 *   pColor->blue
 *   pColor->clear
 *   
 * @return 
 *   EC_SUCCESS if it was successful, an error code otherwise
 */
tError colSensReadRawColor(tColorRawRGB *pColor)
{
  tError error; 
  tCommandReg commandReg; 

  // @ToDo 3: Read ADC Data Registers
  commandReg.Byte = 0;       // Suppress compiler warning C5651;
  commandReg.Bits.CMD = 1;   // Select command register.  Must write as 1.
  commandReg.Bits.Transaction = REG_CMD_TRANS_BLOCK_PROT; // use of Block protocol for I2C
  commandReg.Bits.Address = REG_ADR_DATA_GREEN;           // select register for following read
  
  error = i2cReadCmdData(I2C_COLSENS_ADR, commandReg.Byte, (char*)pColor, sizeof(tColorRawRGB));
  if (error) return error;
 
  // Note the byte ordering in the ADC Channel Data Registers!!
  swapByteOrder(&pColor->red); 
  swapByteOrder(&pColor->green);
  swapByteOrder(&pColor->blue);
    
  return EC_SUCCESS;
}


/**
 * Reads the color as 8-bit RGB value.
 * http://de.wikipedia.org/wiki/RGB-Farbraum
 * 
 * @param [out] pColor
 *   pColor->red [0...255]
 *   pColor->green [0...255]
 *   pColor->blue [0...255]
 *   pColor->clear [0...255]
 *   
 * @return 
 *   EC_SUCCESS if it was successful, an error code otherwise
 */
tError colSensReadColorRGB(tColorRGB *pColor)
{
  tError error; 
  tColorRawRGB c;
  int32 r;
  
  error = colSensReadRawColor(&c);
  if (error) return error;
   
  r = ((int32)c.red - blackValue.red);
  if (r < 0) r = 0;
  else
  {
     r = ((r * 255) / (whiteValue.red - blackValue.red));
     if (r > 255) r = 255;
  }
  pColor->red = (uint8)r;
  
  r = ((int32)c.green - blackValue.green);
  if (r < 0) r = 0;
  else
  {
     r = ((r * 255) / (whiteValue.green - blackValue.green));
     if (r > 255) r = 255;
  }
  pColor->green = (uint8)r;

  r = ((int32)c.blue - blackValue.blue);
  if (r < 0) r = 0;
  else
  {
    r = ((r * 255) / (whiteValue.blue - blackValue.blue));
    if (r > 255) r = 255;
  }
  pColor->blue = (uint8)r;
  
  return EC_SUCCESS;   
}


/**
 * Converts a RGB color into the hsv model
 * http://de.wikipedia.org/wiki/RGB-Farbraum
 * http://de.wikipedia.org/wiki/HSV-Farbraum
 * 
 * @param [in] rgb
 *   the source color (red, green, blue)
 * 
 * @param [out] hsv 
 *   the hsv model of the rgb color
 *   color->saturation [0..1000]
 *   color->value [0..1000]
 *   color->hue [0..360]   
 */
void colSensRGBtoHSV(tColorRGB *rgb, tColorHSV *hsv)
{
  int32 h, min, max;   
  h = 0;
  
  if (rgb->red == rgb->blue && rgb->red == rgb->green)
  {
    h = 0;
    max = min = rgb->red;
  }
  else if ((rgb->red >= rgb->blue) && (rgb->red >= rgb->green))
  {
    max = rgb->red;
    min = ((rgb->blue > rgb->green) ? rgb->green : rgb->blue);
    
    h = (60l * ((int32)rgb->green - rgb->blue)) / (max - min);
  }
  else if ((rgb->green >= rgb->blue) && (rgb->green >= rgb->red))
  {
    max = rgb->green;
    min = ((rgb->blue > rgb->red) ? rgb->red : rgb->blue);
    h = 120l + (60 * ((int32)rgb->blue - rgb->red)) / (max - min);
  }
  else if ((rgb->blue >= rgb->green) && (rgb->blue >= rgb->red))
  {
    max = rgb->blue;
    min = ((rgb->green > rgb->red) ? rgb->red : rgb->green);
    h = 240l + (60 * ((int32)rgb->red - rgb->green)) / (max - min);
  }
  if (h < 0) h += 360;
  
  if (max == 0) hsv->saturation = 0;
  else hsv->saturation = (uint16)( (1000*(max-min))/max);  

  hsv->hue = (uint16)h;
  hsv->value = (uint16)((1000 * max)/255);  
}


/**
 * Reads the color und converts it to the HSV-model.
 * http://de.wikipedia.org/wiki/HSV-Farbraum
 * 
 * @param [out] colorHSV
 *   color->saturation [0..1000]
 *   color->value [0..1000]
 *   color->hue [0..360]
 *   
 * @return 
 *   EC_SUCCESS if it was successful, an error code otherwise
 */
tError colSensReadColorHSV(tColorHSV *colorHSV)
{
  tError error;
  tColorRGB colorRGB;
    
  error = colSensReadColorRGB(&colorRGB);
  if (error) return error;
  
  colSensRGBtoHSV(&colorRGB, colorHSV);
      
  return EC_SUCCESS;
}


tError colSensSetWhite(void)
{
  return colSensReadRawColor(&whiteValue);
}


tError colSensSetBlack(void)
{
  return colSensReadRawColor(&blackValue);
}

/**
 * Sets Prescaler Mode and Gain in the Gain Register of TCS3414
 * (Transaction Type = Byte protocol) 
 * 
 * @returns 
 *    EC_SUCCESS or an error code if the device has not responded
 */
tError colSensUpdateGain(void)
{
  tCommandReg commandReg; 

  // @ToDo 2: Configure Gain Register
  commandReg.Byte = 0;       // Suppress compiler warning C5651;
  commandReg.Bits.CMD = 1;   // Select command register.  Must write as 1.
  commandReg.Bits.Transaction = REG_CMD_TRANS_BYTE_PROT; // use of Byte protocol for I2C
  commandReg.Bits.Address = REG_ADR_GAIN;                // select register for following read
  
  return i2cWriteCmdData(I2C_COLSENS_ADR, commandReg.Byte, &gainReg.Byte, sizeof(tControlGainReg));
}


tError colSensWriteStatus(void)
{
  tColorRGB rgb;
  tColorHSV hsv;
  tError error;
  
  error = colSensReadColorRGB(&rgb);
  if (error) return error;
   
  sciWriteLine("color sensor status:");
  sciWrite("  r=");
  termWriteNum16u(rgb.red);
  sciWrite(",g=");
  termWriteNum16u(rgb.green);
  sciWrite(",b=");
  termWriteNum16u(rgb.blue);
  
  colSensRGBtoHSV(&rgb, &hsv);
  sciWrite(",h=");
  termWriteNum16u(hsv.hue);
  sciWrite(",s=");
  termWriteNum16u(hsv.saturation);
  sciWrite(",v=");
  termWriteNum16u(hsv.value);    
  sciWrite("\r\n");
  return EC_SUCCESS;
}


tError colSensParseCommand(char *cmd)
{
  uint8 value;
  if (strcmp(cmd, "help") == 0)
  {
    return EC_SUCCESS;
  }
  else if (strcmp(cmd, "status") == 0)
  {
    return colSensWriteStatus();
  }
  else if (strncmp(cmd, "set ", sizeof("set")) == 0)
  {
    cmd += sizeof("set ") - 1;
    if (*cmd == 'g') // Gain
    { 
      value = (int8)atoi(cmd + 2);
      gainReg.Bits.Gain = value;
      return colSensUpdateGain();
    }
    else if (*cmd == 'p') // Prescaler
    { 
      value = (int8)atoi(cmd + 2);
      gainReg.Bits.Prescaler = value;
      return colSensUpdateGain();      
    }
  }
  return EC_INVALID_CMD;
}


/**
 * Power on and ADC enable through write of appropriate bits in control register 
 * Structure of communication protocol:                                          
 * Start Condition | Byte 1 | Akn | Byte 2 | Akn | Byte 3 | Akn | Stop condition 
 * contents of the transfered data                                               
 * Byte 1          | Byte 2       | Byte 3                                       
 * I2C Address     | Command      | Data                                           
 * Device Adr+W    | commandReg   | Values for ControlStatusReg                  
 * 0x72            | 0x80         | 0x03      
 * 
 * @returns 
 *    EC_SUCCESS or an error code if the device has not responded
 */
tError colSensInit(void)
{
  tError error;
  tCommandReg commandReg;
  tControlStatusReg controlStatusReg;
  
  // set default value
  whiteValue.red = whiteValue.green = whiteValue.blue = 32000;
  blackValue.red = blackValue.green = blackValue.blue = 0;
  
  // enable "under floor lightning"
  PTCDD_PTCDD2 = 1;                 
  PTCD_PTCD2 = 1;
  
  // @ToDo 1: Power on and enable ADC for color sensor TCS3414

  // Power on and enable ADC for color sensor TCS3414
  commandReg.Byte = 0;              // Suppress compiler warning C5651;
  commandReg.Bits.CMD = 1;          // Select command register. Must write as 1.
  commandReg.Bits.Transaction = REG_CMD_TRANS_BYTE_PROT; // use of Byte protocol for I2C
  commandReg.Bits.Address = REG_ADR_CONTROL;             // Select register for following R/W
    
  controlStatusReg.Byte = 0;        // Suppress compiler warning C5651;
  controlStatusReg.Bits.Power = 1;  // power on 
  controlStatusReg.Bits.AdcEn = 1;  // enable ADC
  
  error = i2cWriteCmdData(I2C_COLSENS_ADR, commandReg.Byte, &controlStatusReg.Byte, 1);                             
  if (error) return error;
 
  // configure Gain Register in color sensor TCS3414 to the most sensitive setting
  gainReg.Bits.Gain = 3;
  gainReg.Bits.Prescaler = 0;
  return colSensUpdateGain();
}
