/**
 *--------------------------------------------------------------------\n
 *          HSLU T&A Hochschule Luzern Technik+Architektur            \n
 *--------------------------------------------------------------------\n
 *
 * \brief         PI control
 * \file
 * \author        Christian Jost
 * \date          21.11.2012
 *
 * \b Language:   Ansi-C \n\n
 * \b Target:     MC Car \n
 *
 * $Id: pid.c 414 2012-12-02 14:23:48Z zajost $
 *--------------------------------------------------------------------
 */
#include "platform.h"
#include "motor.h"
#include "pid.h"
#include "quad.h"
#include "sci.h"
#include <string.h>
#include "stdlib.h"
#include "util.h"
#include "term.h"

static tQuadSpeed nomValue; // Istwert

static int16 setValueL, setValueOldL;
static int16 setValueR, setValueOldR;

static uint8 kpL, kiL, kdL;
static uint8 kpR, kiR, kdR;

static int16 integL;
static int16 integR;

static int16 devOldL;
static int16 devOldR;

static uint8 delay=0;
static bool pidEnabled;

static int16 devL, valL;
static int16 devR, valR;

void pidSetSpeedL(int16 speed)
{
  setValueL = speed;
}

void pidSetSpeedR(int16 speed)
{
  setValueR = speed;
}

void pidSetSpeed(int16 speedL, int16 speedR)
{
  setValueL = speedL;
  setValueR = speedR;
}

void pidDoWork(void)
{  
  // readSpeed from Encoder
  if (quadGetSpeed(&nomValue) != EC_SUCCESS) return;
  
  if (delay > 0)
  {
    delay--;
    return;
  }
  
  if (setValueOldL == 0 && setValueL != 0)
  {
    delay = 2;
  }
  setValueOldL = setValueL;
  
  
  // left wheel
  if (setValueL == 0) valL = integL = devOldL = 0;
  else
  {
    // deviation (max devL = +1000 - -1000 = 2000)
    devL = (setValueL - nomValue.speedLeft) / 8;
    
    // P-Part (max kpL = 
    valL = (kpL * devL) / 32;
    
    // I-Part with anti-windup
    if (kiL != 0) integL += devL;    
    valL += (kiL * integL) / 32;
    
    // D-Part
    valL += (kdL*(devL-devOldL)/32);
    devOldL = devL;
    
    // limit control point
    if (valL > 127)
    {
      valL = 127;
      integL -= devL;
    }
    else if (valL < -127)
    {
      valL = -127;
      integL -= devL;
    }
  }

  // right wheel
  if (setValueR == 0) valR = integR = devOldR = 0;
  else
  {
    // control deviation 
    devR = (setValueR - nomValue.speedRight) / 8;
    
    // P-Part
    valR = (kpR * devR) / 32;
    
    // I-Part  with anti-windup
    if (kiR != 0) integR += devR;
    valR += (kiR * integR) / 32;

    // D-Part
    valR += (kdR*(devR-devOldR)/32);
    devOldR = devR;
        
    // limit control point
    if (valR > 127)
    {
      valR = 127;
      integR -= devR;
    }
    else if (valR < -127)
    {
      valR = -127;
      integR -= devR;
    }
  }
  if (pidEnabled)
  {
    motorSetPWMLeft((int8)valL);
    motorSetPWMRight((int8)valR);
  }
}





void pidWriteHelp(void)
{
  sciWriteLine("pid commands:");
  sciWriteLine("  help");
  sciWriteLine("  status");
  sciWriteLine("  set p <value: 0..100>");
  sciWriteLine("  set i <value: 0..100>");
  sciWriteLine("  set d <value: 0..100>");
  sciWriteLine("  set o <0|1>");
}

void pidWriteStatus(void)
{ 
  //sciWrite("  p=23,i=23,d=1,o=1");
  sciWriteLine("pid status:");
  sciWrite("  p=");
  termWriteNum16u(kpL);
  sciWrite(",i=");
  termWriteNum16u(kiL);
  sciWrite(",d=");
  termWriteNum16u(kdL);
  sciWrite(",o=");
  termWriteNum16u(pidEnabled);  
  
  //sciWrite("  L:s=1231,n=2323,d=2323,i=2342,v=123");
  sciWrite("\n  L:s=");
  termWriteNum16s(setValueL);
  sciWrite(",n=");
  termWriteNum16s(nomValue.speedLeft);
  sciWrite(",d=");
  termWriteNum16s(devL);
  sciWrite(",i=");
  termWriteNum16s(integL);
  sciWrite(",v=");
  termWriteNum16s(valL);
  
  //sciWrite("  R:s=1231,n=2323,d=2323,i=2342,v=123");
  sciWrite("\n  R:s=");
  termWriteNum16s(setValueR);
  sciWrite(",n=");
  termWriteNum16s(nomValue.speedRight);
  sciWrite(",d=");
  termWriteNum16s(devR);
  sciWrite(",i=");
  termWriteNum16s(integR);
  sciWrite(",v=");
  termWriteNum16s(valR);
  sciWrite("\n");
}

tError pidParseCommand(char *cmd)
{
  uint8 value;
  if (strcmp(cmd, "help") == 0)
  {
    pidWriteHelp();
    return EC_SUCCESS;
  }
  else if (strcmp(cmd, "status") == 0)
  {
    pidWriteStatus();
    return EC_SUCCESS;
  }
  else if (strncmp(cmd, "set ", sizeof("set")) == 0)
  {
    cmd += sizeof("set ") - 1;
    if (*cmd == 'p')
    { 
      value = (int8)atoi(cmd + 2);
      if (value <= 100)
      {
        kpL = kpR = value;
        return EC_SUCCESS;
      }      
    }
    else if (*cmd == 'i')
    { 
      value = (int8)atoi(cmd + 2);
      if (value <= 100)
      {
        kiL = kiR = value;
        return EC_SUCCESS;
      }      
    }
    else if (*cmd == 'd')
    { 
      value = (int8)atoi(cmd + 2);
      if (value <= 100)
      {
        kdL = kdR = value;
        return EC_SUCCESS;
      }      
    }
    else if (*cmd == 'o')
    { 
      pidEnabled = (int8)atoi(cmd + 2);
      return EC_SUCCESS;
    } 
  }
  return EC_INVALID_CMD;
}


void pidInit(void)
{
  setValueL = setValueR = 0;
  setValueOldL = setValueOldR = 0;
  
  integL = integR = 0;
  devOldL = devOldR = 0;
  
  pidEnabled = TRUE;
  
  kiL = kiR = 12; // max 20 
  kpL = kpR = 50; // max 128
  kdL = kdR = 1; // max 40
}